/*
 * USB-MIDI-Controller für Arduino Pro Micro (ATmega32u4)
 * 
 * Hardware:
 * - 2x MCP23017 I/O Expander (I2C)
 * - 1x CD74HC4067 Multiplexer (16-Kanal, analog)
 * - 8x Drehpotis + 4x Schieberegler (12 analoge Regler total)
 * - 20x Buttons (Tact Switches)
 * - 20x LEDs (optisches Feedback)
 * 
 * Funktionen:
 * - Analoge Regler senden MIDI-CC (CC1-CC12)
 * - Buttons senden MIDI Note-On/Off
 * - LEDs werden via eingehende MIDI-Nachrichten angesteuert
 * - USB-MIDI Kommunikation
 */

#include <Arduino.h>
#include <Wire.h>
#include <Adafruit_MCP23X17.h>

// Eigene Module
#include "config.h"
#include "multiplexer.h"
#include "button_handler.h"
#include "led_controller.h"
#include "midi_handler.h"

// Hardware-Objekte
Adafruit_MCP23X17 mcp1; // Erste MCP23017 (Adresse 0x20)
Adafruit_MCP23X17 mcp2; // Zweite MCP23017 (Adresse 0x21)

// Modul-Objekte
Multiplexer mux(MUX_S0, MUX_S1, MUX_S2, MUX_S3, MUX_ANALOG_PIN);
ButtonHandler buttons(&mcp1, &mcp2);
LEDController leds(&mcp1, &mcp2);

// Variablen für Timing
unsigned long lastAnalogRead = 0;
unsigned long lastButtonRead = 0;
unsigned long lastMIDIRead = 0;

// Timing-Intervalle (in Milliseconds)
const unsigned long ANALOG_READ_INTERVAL = 10;  // 100Hz für analoge Regler
const unsigned long BUTTON_READ_INTERVAL = 5;   // 200Hz für Buttons (responsive)
const unsigned long MIDI_READ_INTERVAL = 1;     // 1000Hz für MIDI (sehr responsive)

// Callback-Funktionen für eingehende MIDI-Nachrichten
void onMIDINoteOn(uint8_t note, uint8_t velocity) {
  // Suche die entsprechende LED für diese Note
  for(int i = 0; i < NUM_BUTTONS; i++) {
    if(BUTTON_NOTE_MAP[i] == note) {
      leds.setLED(i, true);
      Serial.print("LED ");
      Serial.print(i);
      Serial.print(" eingeschaltet für Note ");
      Serial.println(note);
      break;
    }
  }
}

void onMIDINoteOff(uint8_t note, uint8_t velocity) {
  // Suche die entsprechende LED für diese Note
  for(int i = 0; i < NUM_BUTTONS; i++) {
    if(BUTTON_NOTE_MAP[i] == note) {
      leds.setLED(i, false);
      Serial.print("LED ");
      Serial.print(i);
      Serial.print(" ausgeschaltet für Note ");
      Serial.println(note);
      break;
    }
  }
}

void setup() {
  // Serielle Kommunikation für Debugging
  Serial.begin(115200);
  while(!Serial && millis() < 3000); // Warte max. 3 Sekunden auf Serial
  
  Serial.println("=== USB-MIDI-Controller Start ===");
  Serial.println("Initialisiere Hardware...");
  
  // I2C initialisieren
  Wire.begin();
  Serial.println("I2C Bus initialisiert");
  
  // MCP23017 Chips initialisieren
  if (!mcp1.begin_I2C(MCP1_ADDRESS)) {
    Serial.println("FEHLER: MCP23017 #1 nicht gefunden!");
    while(1); // Stoppe hier bei Fehler
  }
  Serial.println("MCP23017 #1 (0x20) initialisiert");
  
  if (!mcp2.begin_I2C(MCP2_ADDRESS)) {
    Serial.println("FEHLER: MCP23017 #2 nicht gefunden!");
    while(1); // Stoppe hier bei Fehler
  }
  Serial.println("MCP23017 #2 (0x21) initialisiert");
  
  // Module initialisieren
  mux.init();
  buttons.init();
  leds.init();
  MIDIHandler::init();
  
  // MIDI-Callbacks registrieren
  MIDIHandler::onNoteOnCallback = onMIDINoteOn;
  MIDIHandler::onNoteOffCallback = onMIDINoteOff;
  
  // LED-Test: Alle LEDs kurz blinken lassen
  Serial.println("LED-Test...");
  for(int i = 0; i < NUM_LEDS; i++) {
    leds.setLED(i, true);
    delay(50);
    leds.setLED(i, false);
  }
  
  Serial.println("=== Initialisierung abgeschlossen ===");
  Serial.println("MIDI-Controller bereit!");
  Serial.println("");
}

void loop() {
  unsigned long currentTime = millis();
  
  // 1. Analoge Regler verarbeiten (Potis & Schieberegler)
  if(currentTime - lastAnalogRead >= ANALOG_READ_INTERVAL) {
    lastAnalogRead = currentTime;
    
    for(int i = 0; i < NUM_ANALOG_CONTROLS; i++) {
      if(mux.hasChanged(i, ANALOG_THRESHOLD)) {
        // Lese den aktuellen Wert
        int rawValue = mux.readChannel(i);
        
        // Konvertiere zu MIDI-Wert (0-127)
        uint8_t midiValue = map(rawValue, 0, 1023, 0, 127);
        
        // Sende MIDI-CC
        MIDIHandler::sendControlChange(ANALOG_CC_MAP[i], midiValue);
      }
    }
  }
  
  // 2. Buttons verarbeiten
  if(currentTime - lastButtonRead >= BUTTON_READ_INTERVAL) {
    lastButtonRead = currentTime;
    
    // Button-Zustände aktualisieren
    buttons.update();
    
    // Prüfe auf Button-Events
    for(int i = 0; i < NUM_BUTTONS; i++) {
      if(buttons.wasPressed(i)) {
        // Button wurde gedrückt - sende Note-On
        MIDIHandler::sendNoteOn(BUTTON_NOTE_MAP[i], 127);
      }
      else if(buttons.wasReleased(i)) {
        // Button wurde losgelassen - sende Note-Off
        MIDIHandler::sendNoteOff(BUTTON_NOTE_MAP[i], 0);
      }
    }
  }
  
  // 3. Eingehende MIDI-Nachrichten verarbeiten
  if(currentTime - lastMIDIRead >= MIDI_READ_INTERVAL) {
    lastMIDIRead = currentTime;
    MIDIHandler::processIncomingMIDI();
  }
}